## util
import os
import logging
import argparse
from tqdm import tqdm_notebook as tqdmnb
from tqdm import tqdm as tqdm
import pickle
import json 
import jsonlines as jsonl
from collections import defaultdict
from typing import Iterable, List, Dict, Tuple, Union
from pathlib import Path
## Stanza
import stanza
from stanza.models.common.doc import Document as StanzaDocument
from stanza.pipeline.core import Pipeline as StanzaPipeline
# self

logger = logging.getLogger(__name__)

# ANLI
anli_p_field = "context"
anli_h_field = "hypothesis"
anli_l_field = "label"
anli_id_field = "uid"

# this is the fields (in, out) called by the program
in_fields = [
    "sentence1", # premise 
    "sentence2", # hypothesis
    "gold_label", # label here
    "pairID", # problem id here
]
out_fields = ["sentence1", "sentence2", "gold_label", "id"]
# for HANS, there is only two labels
labels = ["non-entailment", "entailment"]
# for MNLI+HANS, may have to do some trick in the dataset reader (HANSreader)
# labels = ["n", "e", "c"]

"""
defaults usage
python preprocess.py -i "in" -o "out" -m $max_num_of_instances -files "filelist" -f(if want to force exe)
ex:
python preprocess.py -i ../data/anli_v1.0/R2/ \
 -o ../data/anli_v1.0_preprocessed/R2/ \
 --files dev.jsonl test.jsonl train.jsonl \
 --force \
 -m 10
 
src# python preprocess.py -i ../data/anli_v1.0/R1/  -o ../data/anli_v1.0_preprocessed/R1/  --files dev.jsonl test.jsonl train.jsonl  --force 

"""


# process one data
def process_one(data :Dict, parser: StanzaPipeline, f: List, out_f: List) -> Dict:
    ret = {}
    ret[out_f[0]] = parser(data[f[0]]).to_dict()
    ret[out_f[1]] = parser(data[f[1]]).to_dict()
    ret[out_f[2]] = data[f[2]]
    ret[out_f[3]] = data[f[3]]
    return ret

# parse MNLI style dataset with Stanza and save the result
def process_file(data_file : Union[str, Path],
                 target_file : Union[str, Path],
                 parser : StanzaPipeline,
                 input_fields: List,
                 output_fields: List,
                 max_num_of_instances: int = None,
                 force_exe : bool=False) -> Iterable[Dict]:
    """
    input
        (data_file = str, target_file = str)
    effect
        load preprocess models
        preprocess and save data to target_file
    ouput
        preprocessed data
    
    parsed data is in jsonl (each line is a json)
    {
        "id" : id: str
        out_f[0] : p doc: StanzaDoc
        out_f[1] : h doc: StanzaDoc
        out_f[2] : label: str
    }
    """
    # alias
    in_f = input_fields
    out_f = output_fields
    # data_file_loading
    with open(data_file, "r") as fo:
        json_data = [json.loads(line) for line in fo]
        
    if max_num_of_instances is not None:
        logger.info(msg=f"max_num_of_instances is {max_num_of_instances}")
        json_data = json_data[:max_num_of_instances]
        
    # check if already processed
    if os.path.isfile(str(target_file)) and not force_exe:
        logger.critical(msg=f"file {target_file}  already exist")
        logger.critical(msg="if u still want to procceed, add force_exe=True in arg")
        return None
    else:
        logger.info(msg=f"creating {target_file} to save result")
        
    # processing and jsonl saving
    with jsonl.open(target_file, mode='w') as writer:
        parsed_data = []
        for data in tqdm(json_data):
            # only add those who have gold labels
            if(data[input_fields[2]] not in labels):
                continue
            pdata = process_one(data, parser, input_fields, output_fields)
            parsed_data.append(pdata)
            writer.write(pdata)
            
    logger.info(msg=f"parsing compplete, data saved to {target_file}")
    return parsed_data

def init_args(arg_string=None):
    parser = argparse.ArgumentParser(description='Preprocess datasets. Note that you should configure your field names and labels in this preprocess.py')
    parser.add_argument('-i', '--input_dir', dest='input_dir', type=str, default=None, required=True, help='input dir should end with / ')
    parser.add_argument('-o', '--output_dir', dest='output_dir', type=str, default=None, required=True)
    parser.add_argument('--files', dest='files', nargs='+', type=str, default=None, required=True)
    parser.add_argument('-m', '--max_num_of_instances', type=int, default=None)
    parser.add_argument('-f', '--force', action='store_true')
    args = parser.parse_args(arg_string)
    return args
               
if __name__ == "__main__":
    args = init_args()
    # handle directory
    if not os.path.isdir(args.input_dir):
        raise InputDirNotExistError
    if os.path.isdir(args.output_dir) and not args.force:
        raise OutputDirExistError
    if not os.path.exists(args.output_dir):
        logger.info(f"making directory {args.output_dir}")
        os.makedirs(args.output_dir) # already checked force
    else:
        logger.info(f"{args.output_dir} exists but forcing execution")
    # parser
    stanza.download('en')
    nlp = stanza.Pipeline(lang='en', processors='tokenize,mwt,pos,lemma,depparse')
    
    # process files
    for file in args.files:
        if not os.path.isdir(args.input_dir):
            raise InputFileNotExistError
        process_file(data_file=args.input_dir+file,
                     target_file=args.output_dir+file,
                     parser = nlp,
                     max_num_of_instances=args.max_num_of_instances,
                     input_fields=in_fields,
                     output_fields=out_fields,
                     force_exe=args.force)


from allennlp.common import Registrable
from typing import Union, List

import torch

from overrides import overrides

class NodeUpdater(torch.nn.Module, Registrable):
    """
    cross attention  for sparse batches 
    """
    pass


@NodeUpdater.register("gru", exist_ok=True)
class GRUNodeUpdater(NodeUpdater):
    """
    use previous node state as hidden,
    the next node state is GRU(message, hidden)
    message can be list, can use concat to aggregate
    """
    
    def __init__(self,
                 input_size: int,  # dim_encoder + dim_attention
                 hidden_size: int, # din_encoder
                 bias: bool = True,
                 aggr: str = "concat", # concat or sum
                ):
        super(GRUNodeUpdater, self).__init__()
        self._rnn = torch.nn.GRUCell(input_size, hidden_size, bias)
        
        assert(aggr in ["concat", "sum"])
        if aggr == "concat":
            self._aggr = torch.cat
        elif aggr == "sum":
            self._aggr = torch.sum
            
    def forward(self,
                message: Union[List[torch.Tensor], torch.Tensor],
                hidden: torch.Tensor,
               ):
        msg = self._aggr(message, dim=-1) # concat the representation dim (i.e. -1)
        return self._rnn(msg, hidden)


from overrides import overrides
from typing import Dict, Tuple, List

import torch
from allennlp.modules import MatrixAttention

from src.tensor_op import sorted_dynamic_parition, dense2sparse, sparse2dense
from .graph_pair2graph_pair_encoder import GraphPair2GraphPairEncoder

# from src.nn.functional import vec_diff


@GraphPair2GraphPairEncoder.register("att_diff", exist_ok=True)
class AttentiveSumDiff(GraphPair2GraphPairEncoder):
    """
    `AttentiveSumDiff` is a `GraphPair2GraphPairEncoder` that produce compmarison between graph pair by,
    Attentive Pooling with similarity function from another graph (AttentiveSum) + 
    Comparison (Diff),
    A naive implementation is from `https://arxiv.org/pdf/1904.12787.pdf`,
    where consine similarity and substraction is used
    
    dynamic_partition + cross attention + substraction
    dim_match = dim_encode
    # can add a projection before output, but this should be left for the NodeUpdater part
    """
    def __init__(self,
                 att: MatrixAttention,
                 dim: int,
                ):
        super(AttentiveSumDiff, self).__init__()
        self._att = att
        self._dim_match = dim
    
    def get_output_dim():
        return self._dim_match
    
    def forward(self,
               x1: torch.Tensor,
               x2: torch.Tensor,
               b1: torch.Tensor,
               b2: torch.Tensor,
               return_attention: bool = False) -> Tuple[torch.Tensor]:
        # N is number of batches
        N: int = b1[-1]+1
        xs1: List[torch.Tensor] = sorted_dynamic_parition(x1, b1)
        xs2: List[torch.Tensor] = sorted_dynamic_parition(x2, b2)
        # return all nodes
        r1 = []
        r2 = []
        Ss = []
        # for each instance in batch
        for i in range(N):
            bx1 = xs1[i] # n*d
            bx2 = xs2[i] # m*d
            S = self._att(bx1.unsqueeze(0), bx2.unsqueeze(0)).squeeze(0) # n*m similarity matrix
            Ss += [S.detach().T]
            r1 += [torch.mm(S, bx2)] # n*m, m*d => n*d, content is from bx2
            r2 += [torch.mm(S.T, bx1)] # m*n, n*d => m*d, content is from bx1
        # concat to make sparse batch
        r1 = torch.cat(r1, dim=0) # 
        r2 = torch.cat(r2, dim=0) # 
        # calculate the difference (use simple substraction here)
        d1 = x1-r1
        d2 = x2-r2
        # return attention if required, TODO
        if return_attention is True:
            return d1, d2, Ss
        else:
            return d1, d2
    
    def __repr__(self):
        return f"({str(self._att)}, dim_matching_output={str(self._dim_match)})"
   


"""
a allennlp `Predictor` for generating HANS prediction result 
which combines neutal and contradiction label to a single `non-entailment` label
so that the ouput file can be processed by hans repo's `evaluate_heur_output.py`

plan to use a predictor 
but ipython notebook may be good enough 
by
for i in range(len(dev_set)):
    res = model(**batch)
    ...
"""


"""
https://github.com/allenai/allennlp/blob/master/allennlp/modules/seq2seq_encoders/pytorch_seq2seq_wrapper.py
"""

from src.modules.graph2graph_encoders.graph2graph_encoder import Graph2GraphEncoder 


from typing import Callable

import torch
import torch_geometric ## nn __init__ has from .conv import * can use directly
from overrides import overrides

from allennlp.common import Registrable

from .RGCNConv import RGCNConv
from .HGTConv import HGTConv
from .GATConv import GATConv
from .CGConv import CGConv

"""
An `Graph2GraphEncoder` is known as `GraphConvolutionLayer`
that takes some graph input and return its updated version.
For the most part we just use
[PyTorch Geometric GraphConvs](https://pytorch-geometric.readthedocs.io/en/latest/modules/nn.html).
Here we provide a thin wrapper to allow registering them and instantiating them `from_params`.
The available options are
* ["gat"](link_to_gat)
* ["rgcn"](link_to_rgcn)
"""

"""
https://github.com/allenai/allennlp/blob/master/allennlp/modules/seq2seq_encoders/pytorch_seq2seq_wrapper.py
"""

class Graph2GraphEncoder(torch.nn.Module, Registrable):
    """
    A `Graph2GraphEncoder` is a `Module` that
    takes a graph and returs its updated version
    
    the API to outside world are `self.in_channels`, `self.out_channels`
    
    Input shape :
        `(SparseBatch, SparseGraphBatch)`;
    output shape:
        `(SparseBatch, SparseGraphBatch)`.
    """
    
    def get_input_dim(self) -> int: # not used
        """
        Returns the dimension of the vector input for each element in the sequence input
        to a `Seq2VecEncoder`.
        """
        raise NotImplementedError
        
    def get_output_dim(self) -> int: # not used
        """
        Returns the dimension of the final vector output by this `Seq2VecEncoder`.  This is `not`
        the shape of the returned tensor, but the last element of that shape.
        """
        raise NotImplementedError


# There are no classes to decorate, so we hack these into Registrable._registry.
# If you want to instantiate it, you can do like this:
# <T:Registrable>.by_name('relu')()
Registrable._registry[Graph2GraphEncoder] = {
    "gat": (GATConv, None),
    #"rgcn": (torch_geometric.nn.conv.RGCNConv, None), 
    "rgcn": (RGCNConv, None),
    "hgt": (HGTConv, None),
    "cg": (CGConv, None),
    # "linear": (lambda: _Graph2GraphEncoderLambda(lambda x: x, "Linear"), None),  # type: ignore
}



from typing import Callable

import torch
import torch_geometric ## nn __init__ has from .conv import * can use directly
from overrides import overrides

from allennlp.common import Registrable

#from .attention import GlobalAttention recursive import


class Graph2VecEncoder(torch.nn.Module, Registrable):
    """
    A `Graph2VecEncoder` is a `Module` that
    takes a graph and returs its pooled representation
    
    Input shape :
        `(SparseBatch of Seq)`;
    output shape:
        `(Batch)`.
    """
    pass

# There are no classes to decorate, so we hack these into Registrable._registry.
# If you want to instantiate it, you can do like this:
# <T:Registrable>.by_name('relu')()
#Registrable._registry[Graph2VecEncoder] = {
#    "global_attention": (GlobalAttention, None), # init takes gate_nn and nn
#    #"linear": (lambda: _Graph2GraphEncoderLambda(lambda x: x, "Linear"), None),  # type: ignore
#}



from setuptools import setup, find_packages

# build long description from README.md
with open("README.md", "r") as fh:
    long_description = fh.read()

listed_requires = [
    # torch
    "torch==1.6.0", #+cu101 is broken
    "torchvision==0.7.0", #+cu101
    # allennlp, should have a lot of dependencies intalled
    "allennlp==1.1.0rc3",
    "allennlp-models==1.1.0rc3",
    # transformers
    "transformers==3.0.2",
    # torch geometric related
    "torch-geometric==1.6.0",
    "torch-scatter==2.0.5",
    "torch-sparse==0.6.7",
    # utils,typing
    "tqdm==4.46.0",
    "typing==3.7.4.3",
    "pathlib",
]   
    
with open('requirements.txt') as fid:
    requires = [line.strip() for line in fid]    
    
setup(
    name="syn_nli", # Replace with your own username
    version="0.0.3", #version
    author="ytlin",
    author_email="0312fs3@gmail.com",
    description="package for the paper Syntax Aware Natural Language Inference@<link>",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EazyReal/2020-IIS-internship",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="allennlp NLP deep learning machine",
    license="Apache",
    packages=find_packages(
        exclude=[
            "*.tests",
            "*.tests.*",
            "tests.*",
            "tests",
            "test_fixtures",
            "test_fixtures.*",
            "benchmarks",
            "benchmarks.*",
            "previous_srcs"
        ]
    ),
    install_requires=listed_requires
    ,
    entry_points={"console_scripts": ["src=src.__main__:run"]},
    include_package_data=True,
    python_requires=">=3.6.1",
    zip_safe=False,
)


"""
inject to allennlp 
"""
# allennlp
from allennlp.modules import FeedForward
from allennlp.common import Registrable

# src
from .graph2vec_encoder import Graph2VecEncoder
from src.tensor_op import sparse2dense

# torch and pytorch geometric
import torch
from torch_scatter import scatter_add
from torch_geometric.utils import softmax

from torch_geometric.nn.inits import reset #modified


@Graph2VecEncoder.register("global_attention", exist_ok=True)
class GlobalAttention(Graph2VecEncoder):
    r"""Global soft attention layer from the `"Gated Graph Sequence Neural
    Networks" <https://arxiv.org/abs/1511.05493>`_ paper
    .. math::
        \mathbf{r}_i = \sum_{n=1}^{N_i} \mathrm{softmax} \left(
        h_{\mathrm{gate}} ( \mathbf{x}_n ) \right) \odot
        h_{\mathbf{\Theta}} ( \mathbf{x}_n ),
    where :math:`h_{\mathrm{gate}} \colon \mathbb{R}^F \to
    \mathbb{R}` and :math:`h_{\mathbf{\Theta}}` denote neural networks, *i.e.*
    MLPS.
    Args:
        gate_nn (torch.nn.Module): A neural network :math:`h_{\mathrm{gate}}`
            that computes attention scores by mapping node features :obj:`x` of
            shape :obj:`[-1, in_channels]` to shape :obj:`[-1, 1]`, *e.g.*,
            defined by :class:`torch.nn.Sequential`.
        nn (torch.nn.Module, optional): A neural network
            :math:`h_{\mathbf{\Theta}}` that maps node features :obj:`x` of
            shape :obj:`[-1, in_channels]` to shape :obj:`[-1, out_channels]`
            before combining them with the attention scores, *e.g.*, defined by
            :class:`torch.nn.Sequential`. (default: :obj:`None`)
    """
    def __init__(self, gate_nn: FeedForward, nn: FeedForward=None):
        super(GlobalAttention, self).__init__()
        self.gate_nn = gate_nn
        self.nn = nn
        self.reset_parameters()

    def reset_parameters(self):
        reset(self.gate_nn)
        reset(self.nn)

    def forward(self, x, batch, size=None, return_attention=False):
        """
        x is node representations
        batch is batch_indices
        size is not used (from original code)
        """
        x = x.unsqueeze(-1) if x.dim() == 1 else x
        size = batch[-1].item() + 1 if size is None else size

        gate = self.gate_nn(x).view(-1, 1)
        x = self.nn(x) if self.nn is not None else x
        assert gate.dim() == x.dim() and gate.size(0) == x.size(0)

        gate = softmax(gate, batch, num_nodes=size)
        out = scatter_add(gate * x, batch, dim=0, dim_size=size)
        
        if return_attention:
            # convert to list of 1*n matrix for visualization
            from src.tensor_op import sorted_dynamic_parition
            return_att = sorted_dynamic_parition(gate, batch)
            return_att = [att.detach().T for att in return_att]
            return out, return_att
        else:
            return out

    def __repr__(self):
        return '{}(gate_nn={}, nn={})'.format(self.__class__.__name__,
                                              self.gate_nn, self.nn)


from src.modules.graph2graph_encoders.graph2graph_encoder import (
    Graph2GraphEncoder,
)
from src.modules.graph_pair2vec_encoders.graph_pair2vec_encoder import (
    GraphPair2VecEncoder,
)
from src.modules.graph2vec_encoders.graph2vec_encoder import (
    Graph2VecEncoder,
)
from src.modules.node_updaters import (
    NodeUpdater,
)

from src.modules.graph_pair2graph_pair_encoders import (
    GraphPair2GraphPairEncoder,
)

from .elmo_token_embedder_with_mask import ElmoTokenEmbedderWithMask


######################################
# must change save_model_folder to designate save param path
# statistics will go to the save_model_folder too
#####################################
import os
from pathlib import Path

TRANSFORMER_NAME = "bert-base-uncased"

######################################
# class for config dict 
#####################################
class Config():
    def __init__(self, *initial_data, **kwargs):
        for dictionary in initial_data:
            for key in dictionary:
                setattr(self, key, dictionary[key])
        for key in kwargs:
            setattr(self, key, kwargs[key])
        return

######################################
# ROOTS 
######################################
SRC_ROOT = Path(os.path.dirname(os.path.realpath(__file__)))
PROJ_ROOT = SRC_ROOT.parent #care if multi source under src

DATA_ROOT = PROJ_ROOT / "data" / "multinli_1.0"
PARAM_PATH = PROJ_ROOT / "param"
P_DATA_ROOT = PROJ_ROOT / "data" / "MNLI_Stanza" # use stanza
PARSER_ROOT = PROJ_ROOT / "parsers"

######################################
# data path
######################################
TEST_FILE = DATA_ROOT / "multinli_1.0_dev_mismatched.jsonl"
DEV_FILE = DATA_ROOT / "multinli_1.0_dev_matched.jsonl"
TRAIN_FILE = DATA_ROOT / "multinli_1.0_train.jsonl"

######################################
# parser gz
######################################
DEP_PARSER_MODEL = PARSER_ROOT / "biaffine-dependency-parser-ptb-2020.04.06.tar.gz"
SRL_LABELER_MODEL = PARSER_ROOT / "bert-base-srl-2020.03.24.tar.gz"

######################################
# parsed data
######################################
P_TEST_FILE = P_DATA_ROOT / "pre_multinli_1.0_dev_mismatched.jsonl"
P_DEV_FILE = P_DATA_ROOT / "pre_multinli_1.0_dev_matched.jsonl"
P_TRAIN_FILE = P_DATA_ROOT / "pre_multinli_1.0_train.jsonl"
P_DEVELOPMENT_FILE = P_DATA_ROOT / "for_development.jsonl"

######################################
# save model
######################################
SAVE_MODEL_FOLDER = "SynNLIv0.1_glove_GAT3"
LOG_FILE_PATH =  PARAM_PATH / SAVE_MODEL_FOLDER / "train_log.txt"
PURE_TRAIN_STAT_PATH = PARAM_PATH / SAVE_MODEL_FOLDER / "stat.jsonl"

######################################
# Preprocssing Config / Data related
######################################
label_to_id = {
    "contradiction" : 0,
    "neutral" : 1,
    "entailment" : 2,
}
id_to_label = ["contradiction", "neutral", "entailment"]

h_field = "sentence2"
p_field = "sentence1"
label_field = "gold_label"
index_field = "pairID"
# alias
hf = "sentence2"
pf = "sentence1"
lf = "gold_label"
idf = "pairID"

######################################
# MODEL
######################################
#CROSS_ATTENTION_HIDDEN_SIZE = 392
NUM_CLASSES = 3
EMBEDDING_D = 300
HIDDEN_SIZE = 300

######################################
# Bert Enbedding
######################################
BERT_EMBEDDING = "bert-base-uncased" #cased?
BERT_MAX_INPUT_LEN = 512
BERT_EMBEDDING_D = 786

######################################
# NLI Model config class and instance
######################################

# for conv encoder
NUM_CONV_LAYERS = 3
NUM_CONV_ATT_HEADS = 3

# for big model
nli_config_dict = {
    "hidden_size" : 300,
    "embedding" : "glove300d",
    "encoder" : "gat",
    "cross_att" : "scaled_dot",
    "aggregation" : "max+mean",
    "local_cmp" : "2-layer-FNN",
    
    "prediction" : "2-layer-FNN",
    "activation" : "relu"
}
model_config = Config(nli_config_dict)

######################################
# Trainning
######################################
"""
reference for bert fine-tuning
BATCH_SIZE = # 32 / 16 / 8
NUM_EPOCHS = # 2 / 3 / 4
LR = # 5/3/2 1e-5
WEIGHT_DECAY = 0.01

reference for KAGNet
BATCH_SIZE = 64
NUM_EPOCHS = 10
LR = 1e-3
"""
trainer_config_dict = {
    "batch_size" : 32,
    "num_epochs" : 5,
    "lr" : 5e-4,
    "weight_decay" : 0.01,
    "max_grad_norm" : 1.0,
    "num_warm_up" : 100,
    "drop_out_prob" : 0.1
}
trainer_config = Config(trainer_config_dict)
BATCH_SIZE = 32
NUM_EPOCHS = 5
LR = 5*1e-4
WEIGHT_DECAY = 0.01
MAX_GRAD_NORM = 1.0
NUM_WARMUP = 100
DROUP_OUT_PROB = 0.1

######################################
# collate
######################################
follow_batch = ["x_p", "x_h", "label"]
tensor_attr_list = ["edge_index_p", "edge_index_h", "x_p", "x_h", "label"]


from src.modules.graph_pair2vec_encoders.graph_pair2vec_encoder import GraphPair2VecEncoder
from src.modules.graph_pair2vec_encoders.graph_embedding_net import GraphEmbeddingNet
from src.modules.graph_pair2vec_encoders.graph_matching_net import GraphMatchingNet


import logging
from typing import Any, Dict, Union
from pathlib import Path

from allennlp.data import Batch, Instance, DatasetReader
from allennlp.common import Registrable
from allennlp.models import Model
from src.interpret import (
    show_matrix_attention,
    show_sequence_attention,
)
from src.data_git.utils import (
    text2graph,
    draw
)

def instance2json(instance: Instance):
    text_p = [token.text for token in  instance.fields["tokens_p"].tokens]
    text_h = [token.text for token in  instance.fields["tokens_h"].tokens]
    gold_label = instance.fields["label"].label
    return {
        "sentence1" : text_p,
        "sentence2" : text_h,
        "gold_label" : gold_label,
    }

class AttentionVisualizer(Registrable):
    """
    visualizer functions
    this tool can be used for error analysis
    """
    # init 
    def __init__(self, model: Model, reader: DatasetReader):
        """
        give a model and reader
        reader with nlp is required for probing
        reader with not be used if provided instance
        """
        self._model = model
        self._reader = reader
        return
    
    # visualize a json dictionary
    def visualize_json(self, json_dict: Dict[str, Any], serialization_dir: Union[str, Path]=None):
        """
        require reader with nlp
        """
        instance = self._reader.text_to_instance(
            premise =  json_dict["sentence1"],
            hypothesis = json_dict["sentence2"],
            gold_label = json_dict["gold_label"] if "gold_label" in json_dict.keys() else None,
        )
        self.visualize_instance(instance, serialization_dir)
        return
        
    # visualize an instance
    def visualize_instance(self, instance: Instance, serialization_dir: Union[str, Path]=None):
        """
        main function of this visualizer
        usage: take an instance and visualize it
        _model have to support "return_attention=True" kwarg
        _model have to have the correct vocab in it
        """
        logger = logging.getLogger(__name__)
        # indexing with model
        instance.index_fields(self._model.vocab)
        # get tokens from instance
        json_dict = instance2json(instance)
        tokens_p = json_dict["sentence1" ]
        tokens_h = json_dict["sentence2" ]
        gold_label = json_dict["gold_label"]
        # get predictions and attentions
        batch = Batch([instance])
        batch_tensor = batch.as_tensor_dict()
        #print(batch.as_tensor_dict())
        ret = self._model.forward(**batch_tensor, return_attention=True)
        ret = self._model.make_output_human_readable(ret)
        pooler_p = ret["attentions"]["pooler1"][0]
        pooler_h = ret["attentions"]["pooler2"][0]
        logger.setLevel(logging.DEBUG)
        logger.info(f"tokens_p are {tokens_p}")
        logger.info(f"tokens_h are {tokens_h}")
        logger.info(f"the predicted label is {ret['predicted_label']}")
        logger.info(f"the gold label is {gold_label}")
        # hope to show_sequence_attention(strlist, att, msg=None)
        if serialization_dir is not None:
            show_sequence_attention(tokens_p, pooler_p, serialization_dir + "pooler_p.png")
            show_sequence_attention(tokens_h, pooler_h, serialization_dir + "pooler_h.png")
            num_layers = 3
            for i in range(num_layers):
                logger.info(f"matrix attention at layer {i}")
                show_matrix_attention(tokens_p, tokens_h, ret["attentions"][f"matching{i}"][0], serialization_dir + f"matching{i}.png")
        else:
            show_sequence_attention(tokens_p, pooler_p)
            show_sequence_attention(tokens_h, pooler_h)
            num_layers = 3
            for i in range(num_layers):
                logger.info(f"matrix attention at layer {i}")
                show_matrix_attention(tokens_p, tokens_h, ret["attentions"][f"matching{i}"][0])
            
    
    def visualize_batch_and_save():
        pass


from typing import Any, Dict
from overrides import overrides

from allennlp.predictors import Predictor
from allennlp.models import Model
from allennlp.common.util import JsonDict, sanitize
from allennlp.data import (
    DatasetReader,
    Instance
)

@Predictor.register("parsed_graph_predictor", exist_ok=True)
class GraphNLIPredictor(Predictor):
    """
    allennlp predictor needs you to overide _json_to_instance
    this predictor takes "parsed" data
    
    Work Flow of Model and Predictor:
    read jsonl file
    self._json_tol_instance
    self._model.forward_on_instances
        self._model.make_output_human_readable
    return result
    """
    
    def __init__(self,
                 model: Model,
                 dataset_reader: DatasetReader,
                 frozen: bool = True
                ) -> None:
        if frozen:
            model.eval()
        self._model = model
        self._dataset_reader = dataset_reader
        self.cuda_device = next(self._model.named_parameters())[1].get_device()
        
    @overrides
    def _json_to_instance(self, json_dict: JsonDict) -> Instance:
        """
        Converts a JSON object into an [`Instance`]
        and a `JsonDict` of information which the `Predictor` should pass through,
        such as tokenised inputs.
        
        dependenct on nlp
        """
        instance = self._dataset_reader.text_to_instance(
            premise =  json_dict["sentence1"],
            hypothesis = json_dict["sentence2"],
            gold_label = json_dict["gold_label"] if "gold_label" in json_dict.keys() else None,
        ) 
        return instance